/*
 * GALIGN -- Genius ALIGNment, optimal pairwise sequence alignment by dynamic programming algorithms
 * 
 * Copyright (C) 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>
 *
 * This file is a part of GALIGN.
 *
 * GALIGN is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GALIGN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Keywords: Optimal Pairwise Sequence Alignment
//
// falign52.cpp

#include "bioinf.h"
#include <iostream>
#include <fstream>
using namespace std;

#define STDINOUT "-"
// #define STDINOUT "con"

#define BUFSIZE 20000

#define INFTY 2000000000

#define MATCH   -4
#define INS     -3
#define DEL     -2
#define UNMATCH -1
#define START    0

#define IDENTITYMATCH 6
#define IDENTITYMISMATCH -1

#define MAXFAMSIZE 15000

class SMatrix {
public:
    friend istream & operator>> ( istream &, SMatrix & );
    int operator  (  ) ( int, int ) const;

private:
    int n;
    char Stype;
    int **Stable;

public:
    SMatrix (  ) { Stype = 'I'; }
};

static int buffer[BUFSIZE];

class fastaifstream: public ifstream{
private:
    char h[200];
public:
    fastaifstream( const char *filename ):ifstream(filename){}
    streamsize readFASTA ( int *&x, int &n, char *fasta_header );
};

streamsize fastaifstream::readFASTA( int *&x, int &n, char *fasta_header )
{
    char *s = ( char * ) buffer;

    streamsize gcnt=0;
    *s = '\0';
    while ( !( *s ) ){
        istream::getline ( s, BUFSIZE ); gcnt+=istream::gcount();
        if( istream::eof())
            break;
    }

    if( *s == '>' ){
        for ( n = 0; s[n] && n<199; ++n )
            h[n]=s[n];
        h[n]='\0';
    }
    else
        while(*s) ++s;

    for( n=0; h[n]; ++n )
        fasta_header[n]=h[n];
    fasta_header[n]='\0';


    while( 1 ){
        istream::getline(s, BUFSIZE ); gcnt+=istream::gcount();
        if( *s=='\0' || istream::eof() ){
            h[0]='\0';
            while(*s)  ++s;
            break;
        }
        else if( *s=='>' ){
            for ( n = 0; s[n] && n<199; ++n )
                h[n]=s[n];
            h[n]='\0';
            break;
        }
        while(*s)
            ++s;
    }
    *s='\0';

    s = ( char * ) buffer;
    for ( n = 0; s[n]; ++n );
    n= n<1 ? 1:n;
    x = new int[n];

    int i = n;

    while ( i-- > 0 ){
        char aa=s[i];
        if('a'<=aa && aa<='z')
            aa+='A'-'a';
        x[i] = aa2i ( aa );
    }
    return gcnt;
}

istream & operator>> ( istream & is, SMatrix & s )
{
    is >> s.Stype;
    if ( s.Stype == 'h' ) {
        is >> s.n;
        s.Stable = new int *[s.n];
        int i;

        for ( i = 0; i < s.n; ++i ) {
            s.Stable[i] = new int[i + 1];
            int j;

            for ( j = 0; j <= i; ++j )
                is >> ( s.Stable[i][j] );
        }
    }
    else if ( s.Stype == 'f' ) {
        is >> s.n;
        s.Stable = new int *[s.n];
        int i;

        for ( i = 0; i < s.n; ++i ) {
            s.Stable[i] = new int[s.n];
            int j;

            for ( j = 0; j < s.n; ++j )
                is >> ( s.Stable[i][j] );
        }
    }
    else if ( s.Stype == '@' ) {
        char *filename = ( char * ) buffer;

        is >> filename;
        ifstream ifs ( filename );

        ifs >> s;
        ifs.close (  );
    }
    return is;
}

int SMatrix::operator  (  ) ( int a, int b ) const {
    if ( Stype == 'I' )
        return ( a == b ? IDENTITYMATCH:IDENTITYMISMATCH );
    else if ( Stype == 'h' && a < b )
        return Stable[b][a];
    else
        return Stable[a][b];
}

void globalAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int &S, int *&path, int &L)
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int i, j, ptr;

    f0 = fu = 0;
    for ( j = 0; j < m; ++j )
        F[j] = ( fu += s ( gap_symbol, y[j] ) );

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = f0;
        fu = ( f0 += s ( x[i], gap_symbol ) );
        for ( j = 0; j < m; ++j ) {
            int f;

            S = flu + s ( x[i], y[j] );
            ptr = MATCH;

            f = F[j] + s ( x[i], gap_symbol );
            if ( S < f ) {
                S = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( S < f ) {
                S = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = S;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = n - 1;
    j = m - 1;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        L++;
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }
    L += ( i + j + 2 );

    path = new int[L];

    i = n - 1;
    j = m - 1;
    int r = L;

    while ( i >= 0 && j >= 0 ) {
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }
    while ( i >= 0 ) {
        path[--r] = DEL;
        --i;
    }
    while ( j >= 0 ) {
        path[--r] = INS;
        --j;
    }

    delete F;
    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
}

void exGlobalAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gop, int gex, int &S, int *&path, int &L )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int *D = new int[m];
    int *gapD = new int[m];
    int I, gapI;
    int i, j, ptr;

    f0 = fu = 0;
    for ( j = 0; j < m; ++j ) {
        F[j] = ( fu += ( j > 0 ? gex : gop ) );
        D[j] = -INFTY;
    }

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = f0;
        fu = ( f0 += ( i > 0 ? gex : gop ) );
        I = -INFTY;

        for ( j = 0; j < m; ++j ) {
            int f;

            D[j] += gex;
            (gapD[j])++;
            f = F[j] + gop;
            if ( D[j] < f ){
                D[j] = f;
                gapD[j] = 1;
            }

            I += gex;
            gapI++;
            f = fu + gop;
            if ( I < f ) {
                I = f;
                gapI = 1;
            }

            S = flu + s ( x[i], y[j] );
            ptr = MATCH;

            if ( S < D[j] ) {
                S = D[j];
                ptr = (DEL) ^ ( gapD[j] << 4 );
            }

            if ( S < I ) {
                S = I;
                ptr = (INS) ^ ( gapI << 4 );
            }

            flu = F[j];
            fu = F[j] = S;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = n - 1;
    j = m - 1;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            ++L;
            --i;
            --j;
        }
        else if ( ( ptr | (~15) ) == DEL ) {
            I = ( ( ptr ^ (DEL) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = DEL;
                --i;
            }
        }
        else {                        // if((ptr | (~15))==INS)
            I = ( ( ptr ^ (INS) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = INS;
                --j;
            }
        }
    }
    L += ( i + j + 2 );

    path = new int[L];

    i = n - 1;
    j = m - 1;
    int r = L;

    while ( i >= 0 && j >= 0 ) {
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }
    while ( i >= 0 ) {
        path[--r] = DEL;
        --i;
    }
    while ( j >= 0 ) {
        path[--r] = INS;
        --j;
    }

    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
    delete F;
    delete D;
    delete gapD;
}

void localAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int &S, int *&path, int &L, int &x0, int &y0 )
{
    S = -INFTY;
    int imax, jmax;

    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu;
    int i, j, ptr;

    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = 0;
        fu = 0;
        for ( j = 0; j < m; ++j ) {
            int fmax = 0;

            ptr = START;

            int f = flu + s ( x[i], y[j] );

            // using <= instead of < makes more chance to get more identity in alignment
            if ( fmax <= f ) {
                fmax = f;
                ptr = MATCH;
            }

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            if ( S < fmax ) {
                S = fmax;
                imax = i;
                jmax = j;
            }
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = imax;
    j = jmax;
    L = 0;
    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
        L++;
    }

    path = new int[L];

    i = imax;
    j = jmax;
    int r = L;

    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        x0 = i;
        y0 = j;
        r--;
        path[r] = ptr;
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    delete F;
    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
}

void exLocalAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gop, int gex, int &S, int *&path, int &L, int &x0, int &y0 )
{
    S = -INFTY;
    int imax, jmax;

    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu;
    int *D = new int[m];
    int *gapD = new int[m];
    int I, gapI;
    int i, j, ptr;

    for ( j = 0; j < m; ++j )
        F[j] = D[j] = gapD[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = fu = I = gapI = 0;

        for ( j = 0; j < m; ++j ) {
            int f;

            D[j] += gex;
            (gapD[j])++;
            f = F[j] + gop;
            if ( D[j] < f ) {
                D[j] = f;
                gapD[j] = 1;
            }

            I += gex;
            gapI++;
            f = fu + gop;
            if ( I < f ) {
                I = f;
                gapI = 1;
            }

            int fmax = 0;

            ptr = START;

            f = flu + s ( x[i], y[j] );
            // using <= instead of < makes more chance to get more identity in alignment
            if ( fmax <= f ) {
                fmax = f;
                ptr = MATCH;
            }

            if ( fmax < D[j] ) {
                fmax = D[j];
                ptr = (DEL) ^ ( gapD[j] << 4 );
            }

            if ( fmax < I ) {
                fmax = I;
                ptr = (INS) ^ ( gapI << 4 );
            }

            flu = F[j];
            fu = F[j] = fmax;
            if ( S < fmax ) {
                S = fmax;
                imax = i;
                jmax = j;
            }
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = imax;
    j = jmax;
    L = 0;
    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        if ( ptr == MATCH ) {
            ++L;
            --i;
            --j;
        }
        else if ( ( ptr | (~15) ) == DEL ) {
            I = ( ( ptr ^ (DEL) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = DEL;
                --i;
            }
        }
        else {                        // if((ptr | (~15))==INS)
            I = ( ( ptr ^ (INS) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = INS;
                --j;
            }
        }
    }

    path = new int[L];

    i = imax;
    j = jmax;
    int r = L;

    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        x0 = i;
        y0 = j;
        r--;
        path[r] = ptr;
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
    delete F;
    delete D;
    delete gapD;
}

void repeatedMatch ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int T, int &S, int *&path, int &L )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int f, fmax;
    int i, j, ptr;

    f0 = 0;
    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m + 1] + 1;

        flu = f0;
        ptr = UNMATCH;
        for ( j = 0; j < m; ++j ) {
            f = F[j] - T;
            if ( f0 <= f ) {
                f0 = f;
                ptr = j;
            }
        }
        fu = f0;
        Ptr[i][-1] = ptr;

        for ( j = 0; j < m; ++j ) {
            fmax = f0;
            ptr = UNMATCH;

            f = flu + s ( x[i], y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = MATCH;
            }

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    S = f0;
    int ej = -1;

    for ( j = 0; j < m; ++j ) {
        f = F[j] - T;
        if ( S <= f ) {
            S = f;
            ej = j;
        }
    }

    // tracing path back
    i = n - 1;
    j = ej;
    L = 0;
    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            --i;
            j = ptr;
        }
        else if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS )
            --j;
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        L++;
    }

    path = new int[L];

    i = n - 1;
    j = ej;
    int r = L;

    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            if ( r < L )
                path[r] = ej;
            ej = j;
            --i;
            j = ptr;
            ptr = UNMATCH;
        }
        else if ( ptr == MATCH ) {
            ej = j;
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS ) {
            ej = j;
            --j;
        }
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        path[--r] = ptr;
    }
    path[r] = ej;

    delete F;
    for ( i = 0; i < n; ++i )
        delete ( Ptr[i] - 1 );
    delete Ptr;
}

void overlapAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int &S, int *&path, int &L, int &x0, int &y0)
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, fmax;
    int i, j, ptr;

    S=-INFTY;
    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = fu = 0;
        for ( j = 0; j < m; ++j ) {
            int f;

            fmax = flu + s ( x[i], y[j] );
            ptr = MATCH;

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
        if ( S < fmax ) {
            S = fmax;
            x0 = i;
            y0 = m-1;
        }
    }                             // end for i
    for ( j = 0; j < m; ++j ) {
        if ( S < F[j] ) {
            S = F[j];
            x0 = n-1;
            y0 = j;
        }
    }

    // tracing path back
    i = x0;
    j = y0;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        L++;
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    path = new int[L];

    i = x0;
    j = y0;
    int r = L;
    while ( i >= 0 && j >= 0 ) {
        x0 = i;
        y0 = j;
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    delete F;
    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
}

void exOverlapAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gop, int gex, int &S, int *&path, int &L, int &x0, int &y0 )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, fmax;
    int *D = new int[m];
    int *gapD = new int[m];
    int I, gapI;
    int i, j, ptr;

    S=-INFTY;
    for ( j = 0; j < m; ++j ) {
        F[j] = D[j] = gapD[j] = 0;
    }

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = fu = 0;
        I = gapI = 0;

        for ( j = 0; j < m; ++j ) {
            int f;

            D[j] += gex;
            (gapD[j])++;
            f = F[j] + gop;
            if ( D[j] < f ) {
                D[j] = f;
                gapD[j] = 1;
            }

            I += gex;
            gapI++;
            f = fu + gop;
            if ( I < f ) {
                I = f;
                gapI = 1;
            }

            fmax = flu + s ( x[i], y[j] );
            ptr = MATCH;

            if ( fmax < D[j] ) {
                fmax = D[j];
                ptr = (DEL) ^ ( gapD[j] << 4 );
            }

            if ( fmax < I ) {
                fmax = I;
                ptr = (INS) ^ ( gapI << 4 );
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
        if ( S < fmax ) {
            S = fmax;
            x0 = i;
            y0 = m-1;
        }
    }                             // end for i
    for ( j = 0; j < m; ++j ) {
        if ( S < F[j] ) {
            S = F[j];
            x0 = n-1;
            y0 = j;
        }
    }

    // tracing path back
    i = x0;
    j = y0;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            ++L;
            --i;
            --j;
        }
        else if ( ( ptr | (~15) ) == DEL ) {
            I = ( ( ptr ^ (DEL) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = DEL;
                --i;
            }
        }
        else {                        // if((ptr | (~15))==INS)
            I = ( ( ptr ^ (INS) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = INS;
                --j;
            }
        }
    }

    path = new int[L];

    i = x0;
    j = y0;
    int r = L;
    while ( i >= 0 && j >= 0 ) {
        x0 = i;
        y0 = j;
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
    delete F;
    delete D;
    delete gapD;
}

void overlapRepeatedMatch ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int T, int &S, int *&path, int &L )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int f, fmax;
    int i, j, ptr;

    f0 = 0;
    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m + 1] + 1;

        flu = f0;
        ptr = UNMATCH;
        j = m-1;
        f = F[j] - T;
        if ( f0 <= f ) {
            f0 = f;
            ptr = j;
        }
        fu = f0;
        Ptr[i][-1] = ptr;

        for ( j = 0; j < m; ++j ) {
            fmax = flu + s ( x[i], y[j] );
            ptr = MATCH;

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    S = f0;
    int ej = -1;
    for ( j = 0; j < m; ++j ) {
        f = F[j] - T;
        if ( S <= f ) {
            S = f;
            ej = j;
        }
    }

    // tracing path back
    i = n - 1;
    j = ej;
    L = 0;
    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            --i;
            j = ptr;
        }
        else if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS )
            --j;
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        L++;
    }

    path = new int[L];

    i = n - 1;
    j = ej;
    int r = L;

    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            if ( r < L )
                path[r] = ej;
            ej = j;
            --i;
            j = ptr;
            ptr = UNMATCH;
        }
        else if ( ptr == MATCH ) {
            ej = j;
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS ) {
            ej = j;
            --j;
        }
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        path[--r] = ptr;
    }
    path[r] = ej;

    delete F;
    for ( i = 0; i < n; ++i )
        delete ( Ptr[i] - 1 );
    delete Ptr;
}

void alignSum (  const int *x, int n, const int *y, int m,
                 int *path, int L, int ix , int jy, const SMatrix & s ,
                 int &nI , int &nP , int &nG )
{
    int r, ptr;

    nI = nP = nG = 0;

    for( r = 0; r < L; ++r ) {
        ptr = path[r];
        if ( ptr == MATCH ) {
            if ( x[ix] == y[jy] ) {
                nI++;
            }
            else if ( s ( x[ix], y[jy] ) > 0 ) {
                nP++;
            }
            ix++;
            jy++;
        }
        else if ( ptr == INS ) {
            ++nG;
            jy++;
        }
        else if ( ptr == DEL ) {
            ++nG;
            ix++;
        }
        else if ( ptr == UNMATCH ) {
            ix++;
        }
        else {
            jy = ptr;
            if ( x[ix] == y[jy] ) {
                nI++;
            }
            else if ( s ( x[ix], y[jy] ) > 0 ) {
                nP++;
            }
            ix++;
            jy++;
        }
    }
    nP += nI;
}


void alignDemo ( ostream & os,
                 const int *x, int n, const int *y, int m,
                 char ( *decode ) ( int ), int *path, int L,
                 int x0 = 0, int y0 = 0, const SMatrix & s = SMatrix (  ),
                 bool smart = false, bool annotate = true, int w = 64 )
{
    int r, ptr;
    int ix, jy;
    int r0 = 0;
    int nI, nP, nG;

    alignSum( x, n, y, m, path, L, x0, y0, s, nI, nP, nG );

    if(!smart){
        r0=0;
        while ( r0 < L ) {
            os << ">>>" << endl;
            os.width ( 5 );
            os << x0 + 1 << ' ';
            ix = x0;
            r = r0;
            while ( r < L ) {
                ptr = path[r];
                if ( ptr == INS ) {
                    os << '-';
                }
                else
                    os << decode ( x[ix++] );
                if ( ( ++r ) % w == 0 )
                    break;
            }
            os << endl;
            if ( annotate ) {
                os.width ( 5 );
                os << ' ' << ' ';
                ix = x0;
                jy = y0;
                r = r0;
                while ( r < L ) {
                    ptr = path[r];
                    if ( ptr == MATCH ) {
                        if ( x[ix] == y[jy] )
                            os << '|';
                        else if ( s ( x[ix], y[jy] ) > 0 )
                            os << ':';
                        else
                            os << ' ';
                        ix++;
                        jy++;
                    }
                    else if ( ptr == INS ) {
                        os << ' ';
                        jy++;
                    }
                    else if ( ptr == DEL ) {
                        os << ' ';
                        ix++;
                    }
                    else if ( ptr == UNMATCH ) {
                        os << '#';
                        ix++;
                    }
                    else {
                        jy = ptr;
                        if ( x[ix] == y[jy] ) {
                            os << '|';
                        }
                        else if ( s ( x[ix], y[jy] ) > 0 ) {
                            os << ':';
                        }
                        else
                            os << ' ';
                        ix++;
                        jy++;
                    }
                    if ( ( ++r ) % w == 0 )
                        break;
                }
                os << endl;
            }
            os.width ( 5 );
            os << y0 + 1 << ' ';
            jy = y0;
            r = r0;
            while ( r < L ) {
                ptr = path[r];
                if ( ptr == MATCH || ptr == INS ) {
                    os << decode ( y[jy] );
                    jy++;
                }
                else if ( ptr == DEL )
                    os << '-';
                else if ( ptr == UNMATCH )
                    os << '.';
                else {
                    jy = ptr;
                    os << decode ( y[jy++] );
                }
                if ( ( ++r ) % w == 0 )
                    break;
            }
            os << endl;
            r0 = r;
            x0 = ix;
            y0 = jy;
        }
        os << endl;
    }

    os << "Identities = " << nI << '/' << L << " (" << ( 100 * nI + L / 2 ) / L << "%) ,   ";
    os << "Positives = " << nP << '/' << L << " (" << ( 100 * nP + L / 2 ) / L << "%) ,   ";
    os << "Gaps = " << nG << '/' << (m+n) << " (" << ( 100 * nG + (m+n) / 2 ) / (m+n) << "%)";
}

/* Instance Application */

char mydecode ( int aaa )
{
    return i2nt ( aaa % 20 );
}

int myencode ( char c )
{
    return 17575;
}

int getfamID(const char *h){
    while(*h && (h[0]!=' ' || h[1]!='P' || h[2]!='F'))
        ++h;
    if(*h)
        return atoi(h+3);
    return -1;
}

struct iData{
    int x;
    int i;
};

int compar(const void *p, const void *q){
    return *(int *)q-*(int *)p;
}

int main ( int argc, char *argv[] )
{
    if(parseArgs(argc, argv, "-h")){
        cerr << "falign, Family search by alignment algorithms" << endl;
        cerr << " by Ali Katanforoush, katanfor@ibb.ut.ac.ir, 2005" << endl;
        cerr << "syntax:" << endl;
        cerr << "falign [-c | -s] [-q | -fam | -a | -fc] -i <param_file> -f <sequences_file> [-g <sequences_file_2>]" << endl;
        cerr << "\t -c: output a simple result only contains alignment scores, in columns" << endl;
        cerr << "\t -s: same -c but writes out in a structured format" << endl;
        cerr << "\t By default, a nice view of each alignment would be printed" << endl;
        cerr << "\t -q: search the first sequence against all others" << endl;
        cerr << "\t -fam: search the first sequence of each family with other sequences of the same family" << endl;
        cerr << "\t -a n: alignment for all pairs in <sequence_file> and n top scores will be output" << endl;
        cerr << "\t -fc: search the \"center\" sequence for each family" << endl;
        cerr << "\t By default, if none of options above is specified then a one by two pairwise alignment will be done" << endl;
        cerr << "\t -i file: path of file containing parameters for alignment" << endl;
        cerr << "\t -f file: path of file containing raw or fasta format of sequences" << endl;
        cerr << "\t -g file: the file containing the sequences that the qurey or the first sequence of family would be aligned with them" << endl;
        cerr << "\t By default both input parametrs and sequences would be given from standard console" << endl;
        return 0;
    }

    bool coln=parseArgs(argc, argv, "-c");
    bool smart=( coln || parseArgs(argc, argv, "-s"));
    const char *paramfilename = parseArgsStr ( argc, argv, "-i", "" );
    bool ui=(*paramfilename=='\0');
    ifstream paramin;
    if(ui)
        paramin.open( STDINOUT );
    else
        paramin.open( paramfilename );
    bool pairs=!parseArgs(argc, argv, "-q");
    bool fam=parseArgs(argc, argv, "-fam");
    bool fc=parseArgs(argc, argv, "-fc");
    pairs=(pairs && !fam);
    int ntop=parseArgsInt(argc,argv,"-a");
    bool dosym=(ntop<0);
    if(dosym)
        ntop=-ntop;

    const char *gfilename = parseArgsStr ( argc, argv, "-g", "" );
    bool gf=(*gfilename!='\0');

    if(fc){
        SMatrix s;

        int gop=-11;
        int gex=-1;

        int ( *encode ) ( char );
        char ( *decode ) ( int );
        encode = aa2i;
        decode = i2aa;

        int *path;
        int S;
        int L;

        fastaifstream seqin( parseArgsStr ( argc, argv, "-f", STDINOUT ) );
        int *x[MAXFAMSIZE];
        int n[MAXFAMSIZE];
        char fasta_header[200];
        int famsize=0;
        int famID,famID0;
        famID0=0;
        while(1){
            seqin.readFASTA ( x[famsize], n[famsize], fasta_header );
            if(n[famsize]==1)
                famID=0;
            else
                famID=getfamID(fasta_header);
            if(famID!=famID0 && famID0>0){
                // find the "center" for 0..famsize-1
                int c,i,j,k,l;
                l=0;
                for(i=1;i<famsize;++i)
                    if(n[l]<n[i])
                        l=i;
                int x0, y0;
                int SminL,SminF2,SminF3,Smin;
                SminL=20*n[l];
                k=l;
                for(i=1;i<famsize;++i){
                    exLocalAlignment ( x[l], n[l], x[i], n[i], s, gop, gex, S, path, L, x0, y0 );
                    if(S<SminL){
                        SminL=S;
                        k=i;
                    }
                    delete path;
                }
                SminF2=20*n[k];
                j=k;
                int Sk[MAXFAMSIZE];
                for(i=0;i<famsize;++i){
                    exLocalAlignment ( x[k], n[k], x[i], n[i], s, gop, gex, S, path, L, x0, y0 );
                    if(S<SminF2){
                        SminF2=S;
                        j=i;
                    }
                    Sk[i]=S;
                    delete path;
                }
                SminF3=Smin=20*n[j];
                c=j;
                for(i=0;i<famsize;++i){
                    exLocalAlignment ( x[j], n[j], x[i], n[i], s, gop, gex, S, path, L, x0, y0 );
                    if(S<SminF3)
                        SminF3=S;
                    if(abs(S-Sk[i])<Smin){
                        Smin=S;
                        c=i;
                    }
                    delete path;
                }
                cout << "> PF" << famID0 << 'r' << l << " 1-LONGEST; " << SminL << endl;
                for(i=0;i<n[l];++i)
                    cout << decode(x[l][i]) ;
                cout << endl;
                cout << "> PF" << famID0 << 'r' << k << " 2-FURTHEST; " << SminF2 << endl;
                for(i=0;i<n[k];++i)
                    cout << decode(x[k][i]) ;
                cout << endl;
                cout << "> PF" << famID0 << 'r' << j << " 3-FURTHEST; " << SminF3 << endl;
                for(i=0;i<n[j];++i)
                    cout << decode(x[j][i]) ;
                cout << endl;
                cout << "> PF" << famID0 << 'r' << c << " 4-CENTER; " << Smin << endl;
                for(i=0;i<n[c];++i)
                    cout << decode(x[c][i]) ;
                cout << endl << endl;
                // ................
                delete x[0];
                n[0]=n[famsize];
                x[0]=x[famsize];
                while(--famsize>0)
                    delete x[famsize];
                if(famID==0){
                    delete x[0];
                    break;
                }
            }
            famID0=famID;
            famsize++;
            if(famsize==MAXFAMSIZE-1){
                --famsize;
                delete x[famsize];
            }
        }
        seqin.close();
    }
    else if(ntop){
        int ( *encode ) ( char );
        char ( *decode ) ( int );
        encode = aa2i;
        decode = i2aa;

        SMatrix s;
        if(ui)
            cout << "Enter Scoring Matrix: " << endl;
        paramin >> s;

        int alg;
        if(ui){
            cout << endl;
            cout << "1. Global Alignment" << endl;
            cout << "2. Global Alignment with Affine Gap Penalty" << endl;
            cout << "3. Local Alignment" << endl;
            cout << "4. Local Alignment with Affine Gap Penalty" << endl;
            cout << "5. Alignment with Repeated Matches" << endl;
            cout << "6. Overlap Alignment" << endl;
            cout << "7. Overlap Alignment with Affine Gap Penalty" << endl;
            cout << "8. Overlap Alignment with Repeated Matches" << endl;
            cout << endl;
            cout << "Select a choice: ";
        }
        paramin >> alg;

        int T, gop, gex;
        if ( alg == 5 || alg == 8 ) {
            if(ui)
                cout << "Enter Threshold: ";
            paramin >> T;
        }
        if ( alg == 2 || alg == 4 || alg == 7 ) {
            if(ui)
                cout << "Enter Gap Open Penalty: ( a negative number ) ";
            paramin >> gop;
            if(ui)
                cout << "Enter Gap Extend Penalty: ( a negative number ) ";
            paramin >> gex;
        }
        char SeqType='A';
        const char *seqinFile=parseArgsStr ( argc, argv, "-f", STDINOUT );
        paramin.close();

        char **h=new char *[ntop];
        int **topy=new int *[ntop];
        int *topm=new int[ntop];
        int *topS=new int[ntop];
        int *topL=new int[ntop];
        int **toppath=new int *[ntop];
        int *topx0=new int[ntop];
        int *topy0=new int[ntop];

        char fasta_header_x[200];
        char *fasta_header_y;
        int *x, *y;
        int n, m;
        int S;
        int L;
        int *path;

        fastaifstream seqin( seqinFile );

        streamsize pos,gcnt;
        pos=0;
        while(1){
            gcnt=seqin.readFASTA ( x, n, fasta_header_x );
            if(n==1){
                delete x;
                break;
            }
            fasta_header_x[17]='\0';

            int i,j;
            for(i=0;i<ntop;++i){
                topS[i]=-INFTY;
                topy[i]=new int;
                toppath[i]=new int;
                h[i]=new char;
            }

            fastaifstream seq2( gf ? gfilename : seqinFile );
            if(dosym)
                seq2.ignore(pos);
            pos+=gcnt;
            while(1){
                fasta_header_y=new char[200];
                seq2.readFASTA ( y, m, fasta_header_y );
                if(m==1){
                    delete fasta_header_y;
                    delete y;
                    break;
                }
                int x0, y0;
                x0 = y0 = 0;
                if ( alg == 1 )
                    globalAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L );
                else if ( alg == 2 )
                    exGlobalAlignment ( x, n, y, m, s, gop, gex, S, path, L );
                else if ( alg == 3 )
                    localAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L, x0, y0 );
                else if ( alg == 4 )
                    exLocalAlignment ( x, n, y, m, s, gop, gex, S, path, L, x0, y0 );
                else if ( alg == 5 )
                    repeatedMatch ( x, n, y, m, s, encode ( '-' ), T, S, path, L );
                else if ( alg == 6 )
                    overlapAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L, x0, y0 );
                else if ( alg == 7 )
                    exOverlapAlignment ( x, n, y, m, s, gop, gex, S, path, L, x0, y0 );
                else if ( alg == 8 )
                    overlapRepeatedMatch ( x, n, y, m, s, encode ( '-' ), T, S, path, L );

                j=0;
                for(i=1;i<ntop;++i)
                    if(topS[i]<topS[j])
                        j=i;
                if(S>topS[j]){
                    delete h[j];
                    h[j]=fasta_header_y;
                    delete topy[j];
                    topy[j]=y;
                    topm[j]=m;
                    topS[j]=S;
                    topL[j]=L;
                    delete toppath[j];
                    toppath[j]=path;
                    topx0[j]=x0;
                    topy0[j]=y0;
                }
                else{
                    delete fasta_header_y;
                    delete y;
                    delete path;
                }
            }
            seq2.close();

            iData *iScore=new iData[ntop];
            for(i=0;i<ntop;++i){
                iScore[i].x=topS[i];
                iScore[i].i=i;
            }
            qsort(iScore,ntop,sizeof(iData),compar);
            for(i=0;i<ntop;++i){
                j=iScore[i].i;
                if(topS[j]> -INFTY){
                    cout << fasta_header_x << ' ' << h[j] << ' ';
                    if(!coln) cout << endl;
                    alignDemo ( cout, x, n, topy[j], topm[j], decode, toppath[j], topL[j], topx0[j], topy0[j], s, smart );
                    if(!coln)  cout << endl;
                    else       cout << " ,  ";
                    cout << "Score = " << topS[j] << endl;
                }
            }
            cout << endl;
            delete iScore;
            for(i=0;i<ntop;++i){
                delete h[i];
                delete topy[i];
                delete toppath[i];
            }
            delete x;
        }
        delete topS;
        delete topL;
        delete topm;
        delete topy;
        delete toppath;
        delete h;
        delete topx0;
        delete topy0;
        seqin.close();
    }
    else{
        int *path;
        int ( *encode ) ( char );
        char ( *decode ) ( int );
        int S;
        int L;

        if(ui)
            cout << "Enter Scoring Matrix: " << endl;
        SMatrix s;
        paramin >> s;

        if(ui){
            cout << endl;
            cout << "1. Global Alignment" << endl;
            cout << "2. Global Alignment with Affine Gap Penalty" << endl;
            cout << "3. Local Alignment" << endl;
            cout << "4. Local Alignment with Affine Gap Penalty" << endl;
            cout << "5. Alignment with Repeated Matches" << endl;
            cout << "6. Overlap Alignment" << endl;
            cout << "7. Overlap Alignment with Affine Gap Penalty" << endl;
            cout << "8. Overlap Alignment with Repeated Matches" << endl;
            cout << endl;
            cout << "Select a choice: ";
        }
        int alg;

        paramin >> alg;
        int T, gop, gex;

        if ( alg == 5 || alg == 8 ) {
            if(ui)
                cout << "Enter Threshold: ";
            paramin >> T;
        }
        if ( alg == 2 || alg == 4 || alg == 7 ) {
            if(ui)
                cout << "Enter Gap Open Penalty: ( a negative number ) ";
            paramin >> gop;
            if(ui)
                cout << "Enter Gap Extend Penalty: ( a negative number ) ";
            paramin >> gex;
        }
        char SeqType, cmd;

        SeqType='A';

        encode = aa2i;
        decode = i2aa;

        paramin.close();

        const char *seqinFile=parseArgsStr ( argc, argv, "-f", STDINOUT );
        fastaifstream seqin( seqinFile );
        fastaifstream *seq2=&seqin;
        if(gf)
            seq2=new fastaifstream( gfilename );

        char fasta_header_buf[200], fasta_header_x_buf[200];
        char *fasta_header=fasta_header_buf;
        char *fasta_header_x=fasta_header_x_buf;
        int *x, *y;
        int n, m;
        int famID,famID0;
        famID=0;
        x=NULL;
        while(1){
            if(pairs || !x){
                if(ui)
                    cout << "Enter first sequence: " << endl;

                seqin.readFASTA ( x, n, fasta_header_x );
                if(n==1){
                    delete x;
                    break;
                }
                if(!coln)
                    cout << fasta_header_x << ';' << n << endl;
                else
                    fasta_header_x[17]='\0';
                if(fam)
                    famID=famID0=getfamID(fasta_header_x);
            }

            if(ui)
                cout << "Enter second sequence: " << endl;
            seq2->readFASTA ( y, m, fasta_header );
            if(m==1){
                delete x;
                delete y;
                break;
            }

            if(fam){
                famID=getfamID(fasta_header);
                if(famID!=famID0){
                    cout << endl;
                    delete x;
                    if(gf){
                        if(ui)
                            cout << "Enter first sequence: " << endl;

                        seqin.readFASTA ( x, n, fasta_header_x );
                        if(n==1){
                            delete x;
                            break;
                        }
                        if(!coln)
                            cout << fasta_header_x << ';' << n << endl;
                        else
                            fasta_header_x[17]='\0';
                        if(fam)
                            famID=famID0=getfamID(fasta_header_x);
                    }
                    else{
                        char *temp=fasta_header;
                        fasta_header=fasta_header_x;
                        fasta_header_x=temp;
                        x=y;
                        n=m;
                        if(!coln)
                            cout << fasta_header_x << ';' << n << endl;
                        else
                            fasta_header_x[17]='\0';
                        famID0=famID;
                        continue;
                    }
                }
            }

            if(coln){
                if(famID)
                    cout << famID << ' '  << fasta_header_x << ' '  << fasta_header << ';' << m << ' ';
                else
                    cout << fasta_header_x << ' '  << fasta_header << ';' << m << ' ';
            }
            else
                cout << fasta_header << ';' << m << endl;

            int x0, y0;
            x0 = y0 = 0;
            if ( alg == 1 )
                globalAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L );
            else if ( alg == 2 )
                exGlobalAlignment ( x, n, y, m, s, gop, gex, S, path, L );
            else if ( alg == 3 )
                localAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L, x0, y0 );
            else if ( alg == 4 )
                exLocalAlignment ( x, n, y, m, s, gop, gex, S, path, L, x0, y0 );
            else if ( alg == 5 )
                repeatedMatch ( x, n, y, m, s, encode ( '-' ), T, S, path, L );
            else if ( alg == 6 )
                overlapAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L, x0, y0 );
            else if ( alg == 7 )
                exOverlapAlignment ( x, n, y, m, s, gop, gex, S, path, L, x0, y0 );
            else if ( alg == 8 )
                overlapRepeatedMatch ( x, n, y, m, s, encode ( '-' ), T, S, path, L );

            alignDemo ( cout, x, n, y, m, decode, path, L, x0, y0, s, smart );
            if(!coln)  cout << endl;
            else       cout << " ,  ";
            cout << "Score = " << S << endl;
            if(!coln)  cout << endl;


            delete path;
            if(pairs)  delete x;
            delete y;
        }
        seqin.close();
        if(gf)
            delete seq2;
    }

    return 0;
}
