/*
 * GALIGN -- Genius ALIGNment, optimal pairwise sequence alignment by dynamic programming algorithms
 * 
 * Copyright (C) 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>
 *
 * This file is a part of GALIGN.
 *
 * GALIGN is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GALIGN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Keywords: Optimal Pairwise Sequence Alignment
//
// galign5.cpp

#include "bioinf.h"
#include <iostream>
#include <fstream>
using namespace std;

class SMatrix {
public:
    friend istream & operator>> ( istream &, SMatrix & );
    int operator  (  ) ( int, int ) const;

//private:
public:
    int n;
    char Stype;
    int **Stable;

public:
    SMatrix (  ) { Stype = 'I'; }
};

static int buffer[20000];

#define INFTY 2000000000

#define MATCH   -4
#define INS     -3
#define DEL     -2
#define UNMATCH -1
#define START    0

void readNTSeq ( istream & is, int *&x, int &n )
{
    char *s = ( char * ) buffer;

    *s = '\0';
    while ( !( *s ) )
        is.getline ( s, 19999 );

    for ( n = 0; s[n]; ++n );
    x = new int[n];

    int i = n;

    while ( i-- > 0 )
        x[i] = nt2i ( s[i] );
}

void readAASeq ( istream & is, int *&x, int &n )
{
    char *s = ( char * ) buffer;

    *s = '\0';
    while ( !( *s ) )
        is.getline ( s, 19999 );

    for ( n = 0; s[n]; ++n );
    x = new int[n];

    int i = n;

    while ( i-- > 0 )
        x[i] = aa2i ( s[i] );
}

void readSeq ( istream & is, int *&x, int &n )
{
    int *z = ( int * ) buffer;
    int a;

    is >> a;
    n = 0;
    while ( is && a >= 0 ) {
        z[n++] = a;
        is >> a;
    }
    x = new int[n];

    int i = n;

    while ( i-- > 0 )
        x[i] = z[i];
}

istream & operator>> ( istream & is, SMatrix & s )
{
    is >> s.Stype;
    if ( s.Stype == 'h' ) {
        is >> s.n;
        s.Stable = new int *[s.n];
        int i;

        for ( i = 0; i < s.n; ++i ) {
            s.Stable[i] = new int[i + 1];
            int j;

            for ( j = 0; j <= i; ++j )
                is >> ( s.Stable[i][j] );
        }
    }
    else if ( s.Stype == 'f' ) {
        is >> s.n;
        s.Stable = new int *[s.n];
        int i;

        for ( i = 0; i < s.n; ++i ) {
            s.Stable[i] = new int[s.n];
            int j;

            for ( j = 0; j < s.n; ++j )
                is >> ( s.Stable[i][j] );
        }
    }
    else if ( s.Stype == '@' ) {
        char *filename = ( char * ) buffer;

        is >> filename;
        ifstream ifs ( filename );

        ifs >> s;
        ifs.close (  );
    }
    return is;
}

int SMatrix::operator  (  ) ( int a, int b ) const {
    if ( Stype == 'I' )
        return ( a == b );
    else if ( Stype == 'h' && a < b )
        return Stable[b][a];
    else
        return Stable[a][b];
}

void globalAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int &S, int *&path, int &L)
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int i, j, ptr;

    f0 = fu = 0;
    for ( j = 0; j < m; ++j )
        F[j] = ( fu += s ( gap_symbol, y[j] ) );

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = f0;
        fu = ( f0 += s ( x[i], gap_symbol ) );
        for ( j = 0; j < m; ++j ) {
            int f;

            S = flu + s ( x[i], y[j] );
            ptr = MATCH;

            f = F[j] + s ( x[i], gap_symbol );
            if ( S < f ) {
                S = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( S < f ) {
                S = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = S;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = n - 1;
    j = m - 1;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        L++;
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }
    L += ( i + j + 2 );

    path = new int[L];

    i = n - 1;
    j = m - 1;
    int r = L;

    while ( i >= 0 && j >= 0 ) {
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }
    while ( i >= 0 ) {
        path[--r] = DEL;
        --i;
    }
    while ( j >= 0 ) {
        path[--r] = INS;
        --j;
    }

    delete F;
    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
}

void exGlobalAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gop, int gex, int &S, int *&path, int &L )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int *D = new int[m];
    int *gapD = new int[m];
    int I, gapI;
    int i, j, ptr;

    f0 = fu = 0;
    for ( j = 0; j < m; ++j ) {
        F[j] = ( fu += ( j > 0 ? gex : gop ) );
        D[j] = -INFTY;
    }

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = f0;
        fu = ( f0 += ( i > 0 ? gex : gop ) );
        I = -INFTY;

        for ( j = 0; j < m; ++j ) {
            int f;

            D[j] += gex;
            (gapD[j])++;
            f = F[j] + gop;
            if ( D[j] < f ){
                D[j] = f;
                gapD[j] = 1;
            }

            I += gex;
            gapI++;
            f = fu + gop;
            if ( I < f ) {
                I = f;
                gapI = 1;
            }

            S = flu + s ( x[i], y[j] );
            ptr = MATCH;

            if ( S < D[j] ) {
                S = D[j];
                ptr = (DEL) ^ ( gapD[j] << 4 );
            }

            if ( S < I ) {
                S = I;
                ptr = (INS) ^ ( gapI << 4 );
            }

            flu = F[j];
            fu = F[j] = S;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = n - 1;
    j = m - 1;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            ++L;
            --i;
            --j;
        }
        else if ( ( ptr | (~15) ) == DEL ) {
            I = ( ( ptr ^ (DEL) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = DEL;
                --i;
            }
        }
        else {                        // if((ptr | (~15))==INS)
            I = ( ( ptr ^ (INS) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = INS;
                --j;
            }
        }
    }
    L += ( i + j + 2 );

    path = new int[L];

    i = n - 1;
    j = m - 1;
    int r = L;

    while ( i >= 0 && j >= 0 ) {
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }
    while ( i >= 0 ) {
        path[--r] = DEL;
        --i;
    }
    while ( j >= 0 ) {
        path[--r] = INS;
        --j;
    }

    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
    delete F;
    delete D;
    delete gapD;
}

void localAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int &S, int *&path, int &L, int &x0, int &y0 )
{
    S = 0;
    int imax, jmax;

    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu;
    int i, j, ptr;

    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = 0;
        fu = 0;
        for ( j = 0; j < m; ++j ) {
            int fmax = 0;

            ptr = START;

            int f = flu + s ( x[i], y[j] );

            // using <= instead of < makes more chance to get more identity in alignment
            if ( fmax <= f ) {
                fmax = f;
                ptr = MATCH;
            }

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            if ( S < fmax ) {
                S = fmax;
                imax = i;
                jmax = j;
            }
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = imax;
    j = jmax;
    L = 0;
    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
        L++;
    }

    path = new int[L];

    i = imax;
    j = jmax;
    int r = L;

    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        x0 = i;
        y0 = j;
        r--;
        path[r] = ptr;
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    delete F;
    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
}

void exLocalAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gop, int gex, int &S, int *&path, int &L, int &x0, int &y0 )
{
    S = 0;
    int imax, jmax;

    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu;
    int *D = new int[m];
    int *gapD = new int[m];
    int I, gapI;
    int i, j, ptr;

    for ( j = 0; j < m; ++j )
        F[j] = D[j] = gapD[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = fu = I = gapI = 0;

        for ( j = 0; j < m; ++j ) {
            int f;

            D[j] += gex;
            (gapD[j])++;
            f = F[j] + gop;
            if ( D[j] < f ) {
                D[j] = f;
                gapD[j] = 1;
            }

            I += gex;
            gapI++;
            f = fu + gop;
            if ( I < f ) {
                I = f;
                gapI = 1;
            }

            int fmax = 0;

            ptr = START;

            f = flu + s ( x[i], y[j] );
            // using <= instead of < makes more chance to get more identity in alignment
            if ( fmax <= f ) {
                fmax = f;
                ptr = MATCH;
            }

            if ( fmax < D[j] ) {
                fmax = D[j];
                ptr = (DEL) ^ ( gapD[j] << 4 );
            }

            if ( fmax < I ) {
                fmax = I;
                ptr = (INS) ^ ( gapI << 4 );
            }

            flu = F[j];
            fu = F[j] = fmax;
            if ( S < fmax ) {
                S = fmax;
                imax = i;
                jmax = j;
            }
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    // tracing path back
    i = imax;
    j = jmax;
    L = 0;
    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        if ( ptr == MATCH ) {
            ++L;
            --i;
            --j;
        }
        else if ( ( ptr | (~15) ) == DEL ) {
            I = ( ( ptr ^ (DEL) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = DEL;
                --i;
            }
        }
        else {                        // if((ptr | (~15))==INS)
            I = ( ( ptr ^ (INS) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = INS;
                --j;
            }
        }
    }

    path = new int[L];

    i = imax;
    j = jmax;
    int r = L;

    while ( i >= 0 && j >= 0 && ( ptr = Ptr[i][j] ) != START ) {
        x0 = i;
        y0 = j;
        r--;
        path[r] = ptr;
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
    delete F;
    delete D;
    delete gapD;
}

void repeatedMatch ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int T, int &S, int *&path, int &L )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int f, fmax;
    int i, j, ptr;

    f0 = 0;
    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m + 1] + 1;

        flu = f0;
        ptr = UNMATCH;
        for ( j = 0; j < m; ++j ) {
            f = F[j] - T;
            if ( f0 <= f ) {
                f0 = f;
                ptr = j;
            }
        }
        fu = f0;
        Ptr[i][-1] = ptr;

        for ( j = 0; j < m; ++j ) {
            fmax = f0;
            ptr = UNMATCH;

            f = flu + s ( x[i], y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = MATCH;
            }

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    S = f0;
    int ej = -1;

    for ( j = 0; j < m; ++j ) {
        f = F[j] - T;
        if ( S <= f ) {
            S = f;
            ej = j;
        }
    }

    // tracing path back
    i = n - 1;
    j = ej;
    L = 0;
    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            --i;
            j = ptr;
        }
        else if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS )
            --j;
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        L++;
    }

    path = new int[L];

    i = n - 1;
    j = ej;
    int r = L;

    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            if ( r < L )
                path[r] = ej;
            ej = j;
            --i;
            j = ptr;
            ptr = UNMATCH;
        }
        else if ( ptr == MATCH ) {
            ej = j;
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS ) {
            ej = j;
            --j;
        }
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        path[--r] = ptr;
    }
    path[r] = ej;

    delete F;
    for ( i = 0; i < n; ++i )
        delete ( Ptr[i] - 1 );
    delete Ptr;
}

void overlapAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int &S, int *&path, int &L, int &x0, int &y0)
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, fmax;
    int i, j, ptr;

    S=-INFTY;
    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = fu = 0;
        for ( j = 0; j < m; ++j ) {
            int f;

            fmax = flu + s ( x[i], y[j] );
            ptr = MATCH;

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
        if ( S < fmax ) {
            S = fmax;
            x0 = i;
            y0 = m-1;
        }
    }                             // end for i
    for ( j = 0; j < m; ++j ) {
        if ( S < F[j] ) {
            S = F[j];
            x0 = n-1;
            y0 = j;
        }
    }

    // tracing path back
    i = x0;
    j = y0;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        L++;
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    path = new int[L];

    i = x0;
    j = y0;
    int r = L;
    while ( i >= 0 && j >= 0 ) {
        x0 = i;
        y0 = j;
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    delete F;
    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
}

void exOverlapAlignment ( const int *x, int n, const int *y, int m, const SMatrix & s, int gop, int gex, int &S, int *&path, int &L, int &x0, int &y0 )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, fmax;
    int *D = new int[m];
    int *gapD = new int[m];
    int I, gapI;
    int i, j, ptr;

    S=-INFTY;
    for ( j = 0; j < m; ++j ) {
        F[j] = D[j] = gapD[j] = 0;
    }

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m];

        flu = fu = 0;
        I = gapI = 0;

        for ( j = 0; j < m; ++j ) {
            int f;

            D[j] += gex;
            (gapD[j])++;
            f = F[j] + gop;
            if ( D[j] < f ) {
                D[j] = f;
                gapD[j] = 1;
            }

            I += gex;
            gapI++;
            f = fu + gop;
            if ( I < f ) {
                I = f;
                gapI = 1;
            }

            fmax = flu + s ( x[i], y[j] );
            ptr = MATCH;

            if ( fmax < D[j] ) {
                fmax = D[j];
                ptr = (DEL) ^ ( gapD[j] << 4 );
            }

            if ( fmax < I ) {
                fmax = I;
                ptr = (INS) ^ ( gapI << 4 );
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
        if ( S < fmax ) {
            S = fmax;
            x0 = i;
            y0 = m-1;
        }
    }                             // end for i
    for ( j = 0; j < m; ++j ) {
        if ( S < F[j] ) {
            S = F[j];
            x0 = n-1;
            y0 = j;
        }
    }

    // tracing path back
    i = x0;
    j = y0;
    L = 0;
    while ( i >= 0 && j >= 0 ) {
        ptr = Ptr[i][j];
        if ( ptr == MATCH ) {
            ++L;
            --i;
            --j;
        }
        else if ( ( ptr | (~15) ) == DEL ) {
            I = ( ( ptr ^ (DEL) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = DEL;
                --i;
            }
        }
        else {                        // if((ptr | (~15))==INS)
            I = ( ( ptr ^ (INS) ) >> 4 );
            L += I;
            while ( I-- > 0 ) {
                Ptr[i][j] = INS;
                --j;
            }
        }
    }

    path = new int[L];

    i = x0;
    j = y0;
    int r = L;
    while ( i >= 0 && j >= 0 ) {
        x0 = i;
        y0 = j;
        ptr = path[--r] = Ptr[i][j];
        if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else                        // if(ptr==INS)
            --j;
    }

    for ( i = 0; i < n; ++i )
        delete Ptr[i];
    delete Ptr;
    delete F;
    delete D;
    delete gapD;
}

void overlapRepeatedMatch ( const int *x, int n, const int *y, int m, const SMatrix & s, int gap_symbol, int T, int &S, int *&path, int &L )
{
    int **Ptr = new int *[n];
    int *F = new int[m];
    int fu, flu, f0;
    int f, fmax;
    int i, j, ptr;

    f0 = 0;
    for ( j = 0; j < m; ++j )
        F[j] = 0;

    for ( i = 0; i < n; ++i ) {
        Ptr[i] = new int[m + 1] + 1;

        flu = f0;
        ptr = UNMATCH;
        j = m-1;
        f = F[j] - T;
        if ( f0 <= f ) {
            f0 = f;
            ptr = j;
        }
        fu = f0;
        Ptr[i][-1] = ptr;

        for ( j = 0; j < m; ++j ) {
            fmax = flu + s ( x[i], y[j] );
            ptr = MATCH;

            f = F[j] + s ( x[i], gap_symbol );
            if ( fmax < f ) {
                fmax = f;
                ptr = DEL;
            }

            f = fu + s ( gap_symbol, y[j] );
            if ( fmax < f ) {
                fmax = f;
                ptr = INS;
            }

            flu = F[j];
            fu = F[j] = fmax;
            Ptr[i][j] = ptr;
        }                           // end for j
    }                             // end for i

    S = f0;
    int ej = -1;
    for ( j = 0; j < m; ++j ) {
        f = F[j] - T;
        if ( S <= f ) {
            S = f;
            ej = j;
        }
    }

    // tracing path back
    i = n - 1;
    j = ej;
    L = 0;
    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            --i;
            j = ptr;
        }
        else if ( ptr == MATCH ) {
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS )
            --j;
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        L++;
    }

    path = new int[L];

    i = n - 1;
    j = ej;
    int r = L;

    while ( i >= 0 ) {
        ptr = Ptr[i][j];
        if ( j == -1 ) {
            if ( r < L )
                path[r] = ej;
            ej = j;
            --i;
            j = ptr;
            ptr = UNMATCH;
        }
        else if ( ptr == MATCH ) {
            ej = j;
            --i;
            --j;
        }
        else if ( ptr == DEL )
            --i;
        else if ( ptr == INS ) {
            ej = j;
            --j;
        }
        else if ( ptr == UNMATCH ) {
            j = -1;
            continue;
        }
        path[--r] = ptr;
    }
    path[r] = ej;

    delete F;
    for ( i = 0; i < n; ++i )
        delete ( Ptr[i] - 1 );
    delete Ptr;
}

void alignDemo ( ostream & os,
                 const int *x, int n, const int *y, int m,
                 char ( *decode ) ( int ), int *path, int L, int x0 = 0, int y0 = 0, bool annotate = true, const SMatrix & s = SMatrix (  ), int w = 64 )
{
    int r, ptr;
    int ix, jy;
    int r0 = 0;
    int nI, nP, nG;

    nI = nP = nG = 0;

    while ( r0 < L ) {
        os << ">>>" << endl;
        os.width ( 5 );
        os << x0 + 1 << ' ';
        ix = x0;
        r = r0;
        while ( r < L ) {
            ptr = path[r];
            if ( ptr == INS ) {
                os << '-';
                ++nG;
            }
            else
                os << decode ( x[ix++] );
            if ( ( ++r ) % w == 0 )
                break;
        }
        os << endl;
        if ( annotate ) {
            os.width ( 5 );
            os << ' ' << ' ';
            ix = x0;
            jy = y0;
            r = r0;
            while ( r < L ) {
                ptr = path[r];
                if ( ptr == MATCH ) {
                    if ( x[ix] == y[jy] ) {
                        os << '|';
                        nI++;
                    }
                    else if ( s ( x[ix], y[jy] ) > 0 ) {
                        os << ':';
                        nP++;
                    }
                    else
                        os << ' ';
                    ix++;
                    jy++;
                }
                else if ( ptr == INS ) {
                    os << ' ';
                    jy++;
                }
                else if ( ptr == DEL ) {
                    os << ' ';
                    ix++;
                }
                else if ( ptr == UNMATCH ) {
                    os << '#';
                    ix++;
                }
                else {
                    jy = ptr;
                    if ( x[ix] == y[jy] ) {
                        os << '|';
                        nI++;
                    }
                    else if ( s ( x[ix], y[jy] ) > 0 ) {
                        os << ':';
                        nP++;
                    }
                    else
                        os << ' ';
                    ix++;
                    jy++;
                }
                if ( ( ++r ) % w == 0 )
                    break;
            }
            os << endl;
        }
        os.width ( 5 );
        os << y0 + 1 << ' ';
        jy = y0;
        r = r0;
        while ( r < L ) {
            ptr = path[r];
            if ( ptr == MATCH || ptr == INS ) {
                os << decode ( y[jy] );
                jy++;
            }
            else if ( ptr == DEL ){
                os << '-';
                ++nG;
            }
            else if ( ptr == UNMATCH ) {
                os << '.';
            }
            else {
                jy = ptr;
                os << decode ( y[jy] );
                jy++;
            }
            if ( ( ++r ) % w == 0 )
                break;
        }
        os << endl;
        r0 = r;
        x0 = ix;
        y0 = jy;
    }
    nP += nI;
    os << endl;
    os << "Identities = " << nI << '/' << L << " (" << ( 100 * nI + L / 2 ) / L << "%) ,   ";
    os << "Positives = " << nP << '/' << L << " (" << ( 100 * nP + L / 2 ) / L << "%) ,   ";
    os << "Gaps = " << nG << '/' << (m+n) << " (" << ( 100 * nG + (m+n) / 2 ) / (m+n) << "%)" << endl;
}


/* Instance Application */

char mydecode ( int aaa )
{
    return i2nt ( aaa % 20 );
}

int myencode ( char c )
{
    return 17575;
}

int main ( int argc, char *argv[] )
{
    bool ui=parseArgs(argc, argv, "-i");
    bool pairs=!parseArgs(argc, argv, "-q");

    int *path;
    int ( *encode ) ( char );
    char ( *decode ) ( int );
    int S;
    int L;

    if(ui)
        cout << "Enter Scoring Matrix: " << endl;
    SMatrix s;

    cin >> s;
    if(ui){
        cout << endl;
        cout << "1. Global Alignment" << endl;
        cout << "2. Global Alignment with Affine Gap Penalty" << endl;
        cout << "3. Local Alignment" << endl;
        cout << "4. Local Alignment with Affine Gap Penalty" << endl;
        cout << "5. Alignment with Repeated Matches" << endl;
        cout << "6. Overlap Alignment" << endl;
        cout << "7. Overlap Alignment with Affine Gap Penalty" << endl;
        cout << "8. Overlap Alignment with Repeated Matches" << endl;
        cout << endl;
        cout << "Select a choice: ";
    }
    int alg;

    cin >> alg;
    int T, gop, gex;

    if ( alg == 5 || alg == 8 ) {
        if(ui)
            cout << "Enter Threshold: ";
        cin >> T;
    }
    if ( alg == 2 || alg == 4 || alg == 7 ) {
        if(ui)
            cout << "Enter Gap Open Penalty: ( a negative number ) ";
        cin >> gop;
        if(ui)
            cout << "Enter Gap Extend Penalty: ( a negative number ) ";
        cin >> gex;
    }
    char SeqType, cmd;

    if(ui)
        cout << "Which type of sequence you will enter? [Nucleotide, Aminoacide, Coded] ";
    cin >> SeqType;

    int *x, *y;
    int n, m;
    x=NULL;
    while(1){
        if(pairs || !x){
            if(ui)
                cout << "Enter first sequence: " << endl;

            if ( SeqType == 'N' || SeqType == 'n' ) {
                encode = nt2i;
                decode = i2nt;
                readNTSeq ( cin, x, n );
            }
            if ( SeqType == 'A' || SeqType == 'a' ) {
                encode = aa2i;
                decode = i2aa;
                readAASeq ( cin, x, n );
            }
            if ( SeqType == 'C' || SeqType == 'c' ) {
                encode = myencode;
                decode = mydecode;
                readSeq ( cin, x, n );
            }
            if(n==1){
                delete x;
                break;
            }
        }

        if(ui)
            cout << "Enter second sequence: " << endl;
        if ( SeqType == 'N' || SeqType == 'n' )
            readNTSeq ( cin, y, m );
        if ( SeqType == 'A' || SeqType == 'a' )
            readAASeq ( cin, y, m );
        if ( SeqType == 'C' || SeqType == 'c' )
            readSeq ( cin, y, m );
        if(m==1){
            delete x;
            delete y;
            break;
        }

        int x0, y0;

        x0 = y0 = 0;
        if ( alg == 1 )
            globalAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L );
        else if ( alg == 2 )
            exGlobalAlignment ( x, n, y, m, s, gop, gex, S, path, L );
        else if ( alg == 3 )
            localAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L, x0, y0 );
        else if ( alg == 4 )
            exLocalAlignment ( x, n, y, m, s, gop, gex, S, path, L, x0, y0 );
        else if ( alg == 5 )
            repeatedMatch ( x, n, y, m, s, encode ( '-' ), T, S, path, L );
        else if ( alg == 6 )
            overlapAlignment ( x, n, y, m, s, encode ( '-' ), S, path, L, x0, y0 );
        else if ( alg == 7 )
            exOverlapAlignment ( x, n, y, m, s, gop, gex, S, path, L, x0, y0 );
        else if ( alg == 8 )
            overlapRepeatedMatch ( x, n, y, m, s, encode ( '-' ), T, S, path, L );

        alignDemo ( cout, x, n, y, m, decode, path, L, x0, y0, true, s );
        cout << "Score = " << S << endl << endl;
        delete path;
        if(pairs)  delete x;
        delete y;
    }

    return 0;
}

