/*
 * GALIGN -- Genius ALIGNment, optimal pairwise sequence alignment by dynamic programming algorithms
 * 
 * Copyright (C) 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>
 *
 * This file is a part of GALIGN.
 *
 * GALIGN is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GALIGN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Keywords: Optimal Pairwise Sequence Alignment
//
// bioinf.cpp

#include "bioinf.h"
#include <cstdlib>
using namespace std;

bool parseArgs ( int &argc, argvType & argv, const char *param )
{
    bool r = false;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = true;
            for ( j = i + 1; j < argc; ++j )
                argv[j - 1] = argv[j];
            --argc;
            --i;
        }
    }
    return r;
}

const char *parseArgsStr ( int &argc, argvType & argv, const char *param, const char *dfault )
{
    const char *r = dfault;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = argv[i + 1];
            for ( j = i + 2; j < argc; ++j )
                argv[j - 2] = argv[j];
            argc -= 2;
            i -= 2;
        }
    }
    return r;
}

int parseArgsInt ( int &argc, argvType & argv, const char *param, int dfault )
{
    int r = dfault;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = atoi ( argv[i + 1] );
            for ( j = i + 2; j < argc; ++j )
                argv[j - 2] = argv[j];
            argc -= 2;
            i -= 2;
        }
    }
    return r;
}

double parseArgsNum ( int &argc, argvType & argv, const char *param, double dfault )
{
    double r = dfault;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = atof ( argv[i + 1] );
            for ( j = i + 2; j < argc; ++j )
                argv[j - 2] = argv[j];
            argc -= 2;
            i -= 2;
        }
    }
    return r;
}

int nt2i ( char b )
{
    static int t[26] = { 0, 10, 1, 11, -1, -1, 2, 12, -1, -1, 6, -1, 7,
                         14, -1, -1, -1, 4, 8, 3, 3, 13, 9, 14, 5, -1
                       };
    if ( 'A' <= b && b <= 'Z' )
        return t[b - 'A'];
    else if ( b == '-' )
        return 15;
    else
        return -1;
}

char i2nt ( int i )
{
    static char t[16] = { 'A', 'C', 'G', 'T', 'R', 'Y', 'K', 'M',
                          'S', 'W', 'B', 'D', 'H', 'V', 'N', '-'
                        };
    if ( 0 <= i && i <= 15 )
        return t[i];
    else
        return '?';
}

int ntCompl ( int i )
{
    static char c[16] = { 3, 2, 1, 0, 5, 4, 7, 6, 8, 9, 13, 12, 11, 10, 14, 15 };
    if ( 0 <= i && i <= 15 )
        return c[i];
    else
        return -1;
}

int aa2i ( char a )
{
    static int t[26] = { 0, 20, 4, 3, 6, 13, 7, 8, 9, -1, 11, 10, 12, 2,
                         -1, 14, 5, 1, 15, 16, -1, 19, 17, 22, 18, 21
                       };
    if ( 'A' <= a && a <= 'Z' )
        return t[a - 'A'];
    else if ( a == '*' )
        return 23;
    else if ( a == '-' )
        return 24;
    else
        return -1;
}

char i2aa ( int i )
{
    static char t[25] = { 'A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M',
                          'F', 'P', 'S', 'T', 'W', 'Y', 'V', 'B', 'Z', 'X', '*', '-'
                        };
    if ( 0 <= i && i <= 24 )
        return t[i];
    else
        return '?';
}

const char *aa2AA ( char a )
{
    static const char t[26][4] = { "ALA", "ASX", "CYS", "ASP", "GLU", "PHE", "GLY", "HIS",
                                   "ILE", "???", "LYS", "LEU", "MET", "ASN", "???", "PRO",
                                   "GLN", "ARG", "SER", "THR", "???", "VAL", "TRP", "XAA",
                                   "TYR", "GLX"
                                 };
    if ( 'A' <= a && a <= 'Z' )
        return t[a - 'A'];
    else if ( a == '*' )
        return "END";
    else if ( a == '-' )
        return "GAP";
    else
        return "???";
}

char AA2aa ( char *A )
{
    const char &a = A[0];
    const char &b = A[1];
    const char &c = A[2];

    if ( a == 'A' ) {
        if ( b == 'S' ) {
            if ( c == 'N' )
                return 'N';
            else if ( c == 'P' )
                return 'D';
            else if ( c == 'X' )
                return 'B';
        }
        else if ( b == 'L' && c == 'A' )
            return 'A';
        else if ( b == 'R' && c == 'G' )
            return 'R';
    }
    else if ( a == 'C' ) {
        if ( b == 'Y' && c == 'S' )
            return 'C';
    }
    else if ( a == 'E' ) {
        if ( b == 'N' && c == 'D' )
            return '*';
    }
    else if ( a == 'G' ) {
        if ( b == 'L' ) {
            if ( c == 'N' )
                return 'Q';
            else if ( c == 'U' )
                return 'E';
            else if ( c == 'Y' )
                return 'G';
        }
        else if ( b == 'A' && c == 'P' )
            return '-';
    }
    else if ( a == 'H' ) {
        if ( b == 'I' && c == 'S' )
            return 'H';
    }
    else if ( a == 'I' ) {
        if ( b == 'L' && c == 'E' )
            return 'I';
    }
    else if ( a == 'L' ) {
        if ( b == 'E' && c == 'U' )
            return 'L';
        else if ( b == 'Y' && c == 'S' )
            return 'K';
    }
    else if ( a == 'M' ) {
        if ( b == 'E' && c == 'T' )
            return 'M';
    }
    else if ( a == 'P' ) {
        if ( b == 'H' && c == 'e' )
            return 'F';
        else if ( b == 'R' && c == 'O' )
            return 'P';
    }
    else if ( a == 'S' ) {
        if ( b == 'E' && c == 'R' )
            return 'S';
    }
    else if ( a == 'T' ) {
        if ( b == 'H' && c == 'R' )
            return 'T';
        else if ( b == 'R' && c == 'P' )
            return 'W';
        else if ( b == 'Y' && c == 'R' )
            return 'Y';
    }
    else if ( a == 'V' ) {
        if ( b == 'A' && c == 'L' )
            return 'V';
    }
    else if ( a == 'X' ) {
        if ( b == 'A' && c == 'A' )
            return 'X';
    }
    return '#';
}

int codon2i ( char *c )
{
    static const int codon[64] = { 11, 2, 11, 2, 16, 16, 16, 16, 1, 15, 1, 15, 9, 9, 12, 9, 5, 8, 5,
                                   8, 14, 14, 14, 14, 1, 1, 1, 1, 10, 10, 10, 10, 6, 3, 6, 3, 0, 0, 0,
                                   0, 7, 7, 7, 7, 19, 19, 19, 19, 23, 18, 23, 18, 15, 15, 15, 15,
                                   23, 4, 17, 4, 10, 13, 10, 13
                                 };
    return codon[( ( ( nt2i ( c[0] ) << 2 ) | ( nt2i ( c[1] ) ) ) << 2 ) | ( nt2i ( c[2] ) )];
}

int dih2i ( char *s )
{
    while ( *s == ' ' )
        ++s;
    int i = 180 + atoi ( s );

    if ( *s == '-' )
        i--;
    if ( i < 0 || 360 <= i )
        i = -1;
    return i;
}
